from django.db import models, transaction
from functools import cached_property
import datetime
import json
import pytz

from .exceptions import NoQueueablesException


def utcnow():
    dt = datetime.datetime.now(pytz.utc)
    dt = dt.replace(tzinfo=pytz.utc)
    return dt


class TaskQuerySet(models.QuerySet):
    pass


RUN_TIMEOUT = 60 * 60  # 1 hour
QUEUE_TIMEOUT = 0


class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return obj.isoformat()
        elif isinstance(obj, datetime.timedelta):
            return (datetime.datetime.min + obj).time().isoformat()

        return super(DateTimeEncoder, self).default(obj)


class ResultStatusChoices(models.IntegerChoices):
    NEW = 0
    COMPLETED = 1
    ERRORED = 2


class Task(models.Model):
    id = models.BigAutoField(primary_key=True)
    handler = models.CharField(max_length=255)
    payload = models.JSONField(default=dict, encoder=DateTimeEncoder)
    reattempts = models.IntegerField(default=0)
    run_timeout = models.IntegerField(default=RUN_TIMEOUT)
    queue_timeout = models.IntegerField(default=QUEUE_TIMEOUT)
    created_at = models.DateTimeField(default=utcnow)
    status = models.IntegerField(
        choices=ResultStatusChoices.choices, default=ResultStatusChoices.NEW
    )
    objects = TaskQuerySet.as_manager()

    def __str__(self) -> str:
        name = f"{self.pk}"
        if args := self.payload.get("args"):
            name += f" args: {args}"
        if kwargs := self.payload.get("kwargs"):
            name += f" kwargs: {kwargs}"
        return name

    def __repr__(self) -> str:
        return f"<Task:{self.pk}>"

    def mark_complete(self):
        return Attempt.objects.create(task=self, status=AttemptStatusChoices.COMPLETED)

    def mark_error(self, message=""):
        return Attempt.objects.create(
            task=self, status=AttemptStatusChoices.ERRORED, message=message
        )

    @cached_property
    def last_attempt(self):
        return self.attempt_set.order_by("-created_at").first()


class QueueableQuerySet(models.QuerySet):
    def in_queue(self):
        return self.filter(in_queue=True)

    def in_process(self):
        return self.filter(in_queue=False)

    def get_queueable(self):
        if queueable := self.select_for_update(skip_locked=True).first():
            queueable.in_queue = False
            queueable.save()
            return queueable
        else:
            raise NoQueueablesException


class Queueable(models.Model):
    task = models.OneToOneField(Task, on_delete=models.CASCADE, primary_key=True)
    queue = models.SlugField(default="default")
    key = models.SlugField(unique=True, null=True)
    in_queue = models.BooleanField(default=True)
    priority = models.PositiveSmallIntegerField(default=100)
    objects = QueueableQuerySet.as_manager()

    def __str__(self) -> str:
        return self.task.__str__()

    def __repr__(self) -> str:
        return self.task.__repr__()


class AttemptStatusChoices(models.IntegerChoices):
    COMPLETED = 1
    ERRORED = 2


class Attempt(models.Model):
    task = models.ForeignKey(Task, on_delete=models.CASCADE)
    created_at = models.DateTimeField(default=utcnow)
    status = models.IntegerField(choices=AttemptStatusChoices.choices)
    message = models.TextField(blank=True, default="")
    value = models.JSONField(null=True, encoder=DateTimeEncoder)



from time import sleep
from .models import Attempt


def ping():
    print("ping")


def throw_exception(n=None):
    if hasattr(throw_exception, "_count"):
        throw_exception._count += 1
    else:
        throw_exception._count = 1

    print(throw_exception._count)
    if throw_exception._count == n:
        return None
    raise Exception("This is busted.")


def bad_sql():
    bad_query = Attempt.objects.raw("DELETE FART FROM 12")
    for row in bad_query:
        print(row)


def okay_sql():
    attempt = Attempt.objects.first()
    return True


def sum(x, y):
    print(x + y)
    return x + y


def say(text=None):
    print(text)


def sumsay(x, y, text):
    print(f"{x} + {y} is {text}")


def sleepy():
    sleep(10)



from .handlers import enqueue_callable


class Job:
    @classmethod
    def setup(cls, *args, **kwargs):
        obj = cls(*args, **kwargs)
        return obj

    @classmethod
    def process(cls, *args, **kwargs):
        obj = cls.setup(*args, **kwargs)
        result = obj.run()
        return result

    @classmethod
    def enqueue(cls, *args, **kwargs):
        queueable, task = enqueue_callable(cls.process, args=args, kwargs=kwargs)
        return queueable, task

    def run(self):
        # The main body of the task
        pass


class Sum(Job):
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def run(self):
        return self.x + self.y



from django.db import transaction
from typing import Tuple
import logging

from django_skiptask.models import (
    Task,
    Queueable,
)


logger = logging.getLogger(__name__)


def enqueue_task(
    handler,
    payload=None,
    reattempts=0,
    run_timeout=None,
    queue_timeout=None,
    queue="default",
    key=None,
) -> Tuple[Queueable, Task]:
    if payload is None:
        payload = {}

    with transaction.atomic():
        task_dict = {
            "handler": handler.get_handler_key(),
            "payload": payload,
            "reattempts": reattempts,
        }
        if run_timeout:
            task_dict["run_timeout"] = run_timeout
        if queue_timeout:
            task_dict["queue_timeout"] = queue_timeout

        task = Task.objects.create(**task_dict)
        queueable = Queueable.objects.create(
            task=task,
            queue=queue,
            key=key,
            priority=100,
        )
        return queueable, task



from django.core.management.base import BaseCommand
import signal

from django_skiptask.commands import exit_gracefully
from django_skiptask.worker import Worker


class Command(BaseCommand):
    help = "Starts a worker to process the queue"

    def handle(self, *args, **options):
        signal.signal(signal.SIGINT, exit_gracefully)
        worker = Worker()
        worker.start()



from django.contrib import admin

from . import models

admin.site.register(models.Queueable, admin.ModelAdmin)
admin.site.register(models.Task, admin.ModelAdmin)
admin.site.register(models.Attempt, admin.ModelAdmin)



from concurrent.futures import ThreadPoolExecutor
from django.db import connections
from importlib import import_module
from typing import Tuple
import inspect

from .models import Task, Queueable
from .services import enqueue_task


class HandlerRegistry:
    """Handler registry is a key value store for looking up all registered handlers."""

    def __init__(self):
        self.registry = {}

    def register(self, handler_cls):
        key = handler_cls.get_handler_key()
        self.registry[key] = handler_cls

    def get(self, key):
        return self.registry.get(key, None)


registry = HandlerRegistry()


class Handler:
    @classmethod
    def get_handler_key(cls):
        return f"{cls.__module__}.{cls.__name__}"

    def handle(self, *args, **kwargs):
        raise RuntimeError("Handle not implemented.")

    def enqueue(self, *args, **kwargs):
        raise RuntimeError("Enqueue not implemented. ")


def connection_cleanup(future):
    connections.close_all()


def run_task(task: Task):
    task_args = task.payload.get("args", [])
    task_kwargs = task.payload.get("kwargs", {})

    module = import_module(task.payload["module"])
    name = task.payload["name"]

    if "." in name:
        cls_name, func_name = name.split(".")
        cls = getattr(module, cls_name)
        func_or_class = getattr(cls, func_name)
    else:
        func_or_class = getattr(module, name)

    if inspect.isfunction(func_or_class):
        task_callable = func_or_class
    elif inspect.isclass(func_or_class):
        task_callable = func_or_class.run
    elif inspect.ismethod(func_or_class):
        task_callable = func_or_class
    else:
        raise RuntimeError("Not a callable function or Task class.")

    # Run in a separate thread to ensure db connection isolation between
    with ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(task_callable, *task_args, **task_kwargs)
        future.add_done_callback(connection_cleanup)
        result_value = future.result(timeout=float(task.run_timeout))
    return result_value


def enqueue_callable(
    func_or_class, args=None, kwargs=None, **local_kwargs
) -> Tuple[Queueable, Task]:
    if inspect.isfunction(func_or_class):
        func_or_class_key = func_or_class.__name__
    elif inspect.isclass(func_or_class):
        func_or_class_key = func_or_class.__name__
    elif inspect.ismethod(func_or_class):
        # SomeClass.run
        class_name = func_or_class.__self__.__name__
        func_name = func_or_class.__name__
        func_or_class_key = f"{class_name}.{func_name}"
    else:
        raise RuntimeError("Not a callable function or Task class.")

    module_key = func_or_class.__module__

    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    payload = {
        "module": module_key,
        "name": func_or_class_key,
        "args": args,
        "kwargs": kwargs,
    }
    return enqueue_task(PythonHandler, payload=payload, **local_kwargs)


class PythonHandler(Handler):
    def handle(self, task: Task):
        return run_task(task)

    def enqueue(self, *args, **kwargs):
        return enqueue_callable(*args, **kwargs)


registry.register(PythonHandler)



import django
import multiprocessing
from typing import Callable

from .processing import process_tasks


class Worker:
    def __init__(
        self,
        process_tasks: Callable = process_tasks,
        n_worker_processes: int = 4,
        stop_on_empty_queue: bool = False,
    ) -> None:
        self._process_tasks = process_tasks
        self.n_worker_processes = n_worker_processes
        self.stop_on_empty_queue = stop_on_empty_queue
        self._process_results = []

    def _start_worker_process(self, pool):
        async_result = pool.apply_async(
            self._process_tasks, kwds={"stop_on_empty_queue": self.stop_on_empty_queue}
        )
        return async_result

    def start(self):
        with multiprocessing.Pool(self.n_worker_processes, maxtasksperchild=1, initializer=django.setup) as pool:
            process_results = []
            for i in range(self.n_worker_processes):
                async_result = self._start_worker_process(pool)
                process_results.append(async_result)
            while True:
                for result in process_results:
                    # If a process is done (no matter the reason), maybe start a new one to take its place.
                    if result.ready():
                        process_results.remove(result)
                        # If processing should be done don't launch any more and let things finish up naturally.
                        from django_skiptask import processing

                        if processing.keep_processing:
                            async_result = self._start_worker_process(pool)
                            process_results.append(async_result)
                if not process_results:
                    break

    def stop(self):
        from django_skiptask import processing

        # global keep_processing
        processing.keep_processing = False



from django.db import transaction
from time import sleep
from typing import Tuple, Type
import logging

from .models import (
    Queueable,
    Attempt,
    AttemptStatusChoices,
    ResultStatusChoices,
)

from .exceptions import NoQueueablesException
from .handlers import registry


logger = logging.getLogger(__name__)
keep_processing = True


def process_queueable(queueable):
    task = queueable.task
    logger.info(f"Starting task {task.pk}")

    handler_cls = registry.get(task.handler)
    if not handler_cls:
        raise RuntimeError("No handler found")
    handler = handler_cls()

    try:
        value = handler.handle(task)
        attempt_status = AttemptStatusChoices.COMPLETED
        result_status = ResultStatusChoices.COMPLETED
    except Exception as e:
        logger.info(e)
        value = None
        attempt_status = AttemptStatusChoices.ERRORED
        result_status = ResultStatusChoices.ERRORED
    finally:
        attempt = Attempt.objects.create(task=task, status=attempt_status, value=value)

    task.status = result_status
    task.save()

    if attempt.status == AttemptStatusChoices.COMPLETED:
        logger.info(f"Task {task.pk} completed with attempt {attempt.pk}")
        queueable.delete()
    else:
        logger.info(f"Task {task.pk} failed with attempt {attempt.pk}")
        # Add one to reattempts to count the original attempt
        if task.attempt_set.count() < task.reattempts + 1:
            logger.info(f"Task {task.pk} requeueing")
            queueable.in_queue = True
            queueable.save()


class Processor:
    qs = None
    queue = "default"
    sleep = 2

    def __init__(self, qs=None, stop_on_empty_queue: bool = True) -> None:
        if self.qs is None:
            self.qs = Queueable.objects.filter(queue=self.queue).in_queue()
            self.stop_on_empty_queue = stop_on_empty_queue

    def get_queueable(self):
        with transaction.atomic(durable=True):
            logger.info(f"Checking for queueable")
            queueable = self.qs.get_queueable()
        return queueable

    def process_queueable(self):
        queueable = self.get_queueable()
        process_queueable(queueable)

    def get_sleep(self):
        return self.sleep

    def keep_processing(self):
        return keep_processing

    def process(self):
        while self.keep_processing():
            queueables = True
            try:
                self.process_queueable()
            except NoQueueablesException:
                queueables = False
            if self.stop_on_empty_queue and not queueables:
                return
            elif not queueables:
                # A lil pause before trying to get the next item
                sleep(self.get_sleep())


def process_tasks(
    processor_class: Type[Processor] = Processor, stop_on_empty_queue: bool = True
):
    processor = processor_class(stop_on_empty_queue=stop_on_empty_queue)
    processor.process()



from django.apps import AppConfig


class SkiptaskConfig(AppConfig):
    name = 'django_skiptask'
    default_auto_field = 'django.db.models.AutoField'

    def ready(self):
        from . import handlers



from django.core.management.base import BaseCommand
from time import sleep
import logging
import signal
import sys

from django_skiptask import processing
from django_skiptask.processing import process_tasks

logger = logging.getLogger(__name__)
sigint_happened = False


def exit_gracefully(signum, frame):
    global sigint_happened

    if sigint_happened == True:
        logger.info("Hard exit.")
        sys.exit(1)
    else:
        logger.info("Soft exit. Finishing up current task, then ending.")

    sigint_happened = True
    processing.keep_processing = False

    signal.signal(signal.SIGINT, exit_gracefully)


class SkipTaskCommand(BaseCommand):
    help = 'Starts a worker to process the queue with auto reload'

    def handle(self, *args, **options):
        signal.signal(signal.SIGINT, exit_gracefully)



# Generated by Django 5.2.4 on 2025-07-05 18:45

import django.db.models.deletion
import django_skiptask.models
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Task",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("handler", models.CharField(max_length=255)),
                (
                    "payload",
                    models.JSONField(
                        default=dict, encoder=django_skiptask.models.DateTimeEncoder
                    ),
                ),
                ("reattempts", models.IntegerField(default=0)),
                ("run_timeout", models.IntegerField(default=3600)),
                ("queue_timeout", models.IntegerField(default=0)),
                (
                    "created_at",
                    models.DateTimeField(default=django_skiptask.models.utcnow),
                ),
                (
                    "status",
                    models.IntegerField(
                        choices=[(0, "New"), (1, "Completed"), (2, "Errored")],
                        default=0,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Queueable",
            fields=[
                (
                    "task",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        serialize=False,
                        to="django_skiptask.task",
                    ),
                ),
                ("queue", models.SlugField(default="default")),
                ("key", models.SlugField(null=True, unique=True)),
                ("in_queue", models.BooleanField(default=True)),
                ("priority", models.PositiveSmallIntegerField(default=100)),
            ],
        ),
        migrations.CreateModel(
            name="Attempt",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(default=django_skiptask.models.utcnow),
                ),
                (
                    "status",
                    models.IntegerField(choices=[(1, "Completed"), (2, "Errored")]),
                ),
                ("message", models.TextField(blank=True, default="")),
                (
                    "value",
                    models.JSONField(
                        encoder=django_skiptask.models.DateTimeEncoder, null=True
                    ),
                ),
                (
                    "task",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_skiptask.task",
                    ),
                ),
            ],
        ),
    ]



from django_skiptask.models import (
    Queueable,
    AttemptStatusChoices,
    ResultStatusChoices,
)
from django_skiptask.tasks import ping, throw_exception, bad_sql, okay_sql, sleepy, sum
from django_skiptask.jobs import Sum
from django_skiptask.processing import process_tasks
from django_skiptask.handlers import enqueue_callable
import pytest


@pytest.mark.django_db(transaction=True)
def test_enqueue():
    enqueue_callable(ping)
    assert Queueable.objects.in_queue().count() == 1


@pytest.mark.django_db(transaction=True)
def test_dequeue():
    queueable, task = enqueue_callable(ping)
    assert Queueable.objects.in_queue().count() == 1
    process_tasks(stop_on_empty_queue=True)
    assert Queueable.objects.in_queue().count() == 0
    assert task.attempt_set.count() == 1
    assert task.attempt_set.first().status == AttemptStatusChoices.COMPLETED
    task.refresh_from_db()
    assert task.status == ResultStatusChoices.COMPLETED


@pytest.mark.django_db(transaction=True)
def test_result():
    queueable, task = enqueue_callable(sum, args=(2, 3))
    assert Queueable.objects.in_queue().count() == 1
    process_tasks(stop_on_empty_queue=True)
    assert Queueable.objects.in_queue().count() == 0
    assert task.attempt_set.count() == 1
    assert task.attempt_set.first().status == AttemptStatusChoices.COMPLETED
    assert task.attempt_set.first().value == 5
    task.refresh_from_db()
    assert task.status == ResultStatusChoices.COMPLETED
    assert Queueable.objects.filter(pk=queueable.pk).first() == None


@pytest.mark.django_db(transaction=True)
def test_job_result():
    queueable, task = Sum.enqueue(2, 3)
    assert Queueable.objects.in_queue().count() == 1
    process_tasks(stop_on_empty_queue=True)
    assert Queueable.objects.in_queue().count() == 0
    assert task.attempt_set.count() == 1
    assert task.attempt_set.first().status == AttemptStatusChoices.COMPLETED
    assert task.attempt_set.first().value == 5
    task.refresh_from_db()
    assert task.status == ResultStatusChoices.COMPLETED


@pytest.mark.django_db(transaction=True)
def test_dequeue_two():
    enqueue_callable(ping)
    enqueue_callable(ping)
    assert Queueable.objects.in_queue().count() == 2
    process_tasks(stop_on_empty_queue=True)
    assert Queueable.objects.in_queue().count() == 0


@pytest.mark.django_db(transaction=True)
def test_exception():
    queueable, task = enqueue_callable(throw_exception)
    process_tasks(stop_on_empty_queue=True)
    assert Queueable.objects.in_queue().count() == 0
    assert task.attempt_set.count() == 1
    assert task.attempt_set.first().status == AttemptStatusChoices.ERRORED


@pytest.mark.django_db(transaction=True)
def test_exception_reattempt():
    del throw_exception._count
    queueable, task = enqueue_callable(throw_exception, kwargs={"n": 3}, reattempts=2)
    process_tasks(stop_on_empty_queue=True)
    assert Queueable.objects.in_queue().count() == 0
    assert task.attempt_set.count() == 3
    assert (
        task.attempt_set.order_by("created_at").first().status
        == AttemptStatusChoices.ERRORED
    )
    assert (
        task.attempt_set.order_by("-created_at").first().status
        == AttemptStatusChoices.COMPLETED
    )


@pytest.mark.django_db(transaction=True)
def test_exception_reattempt_fail():
    queueable, task = enqueue_callable(throw_exception)
    process_tasks(stop_on_empty_queue=True)
    assert Queueable.objects.in_queue().count() == 0
    assert task.attempt_set.count() == 1
    assert task.attempt_set.first().status == AttemptStatusChoices.ERRORED
    task.refresh_from_db()
    assert task.status == ResultStatusChoices.ERRORED


@pytest.mark.django_db(transaction=True)
def test_db_transaction_isolation_fail():
    _, bad_task = enqueue_callable(bad_sql)
    _, okay_task = enqueue_callable(okay_sql)
    process_tasks(stop_on_empty_queue=True)
    assert Queueable.objects.in_queue().count() == 0
    # assert task.attempt_set.count() == 1
    assert all(
        [
            x.status == AttemptStatusChoices.ERRORED
            for x in bad_task.attempt_set.all()
        ]
    )
    assert all(
        [
            x.status == AttemptStatusChoices.COMPLETED
            for x in okay_task.attempt_set.all()
        ]
    )
    bad_task.refresh_from_db()
    assert bad_task.status == ResultStatusChoices.ERRORED


@pytest.mark.django_db(transaction=True)
def test_task_timeout():
    _, task = enqueue_callable(sleepy, run_timeout=5)
    process_tasks(stop_on_empty_queue=True)
    assert Queueable.objects.in_queue().count() == 0
    assert task.attempt_set.count() == 1
    assert task.attempt_set.first().status == AttemptStatusChoices.ERRORED
    task.refresh_from_db()
    assert task.status == ResultStatusChoices.ERRORED


# @pytest.mark.django_db()
# def test_processor_system():
#     worker = Worker(stop_on_empty_queue=True)
#     enqueue_callable(ping)
#     enqueue_callable(ping)
#     enqueue_callable(ping)
#     enqueue_callable(ping)
#     skiptask_processing.keep_processing = False
#     worker.start()
#     assert Attempt.objects.count() == 4
#     assert Queueable.objects.in_queue().count() == 0
